#!/usr/bin/env python3
# encoding: utf-8

from __future__ import annotations

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = ["make_application"]

from collections.abc import Callable
from functools import partial
from io import BytesIO
from os import fsdecode, PathLike
from pathlib import Path
from posixpath import dirname, splitext, split as splitpath
from sqlite3 import connect, Connection, OperationalError
from threading import Lock
from typing import Literal
from urllib.parse import quote

from blacksheep import redirect, Application, Request
from blacksheep.client import ClientSession
from blacksheep.server.compression import use_gzip_compression
from blacksheep.server.remotes.forwarding import ForwardedHeadersMiddleware
from blacksheep_client_request import request as blacksheep_request
from encode_uri import encode_uri_component_loose
from p115client import P115Client
from path_predicate import MappingPath
from property import locked_cacheproperty
from urllib3.poolmanager import PoolManager
from uvicorn.middleware.wsgi import WSGIMiddleware
from wsgidav.wsgidav_app import WsgiDAVApp # type: ignore
from wsgidav.dav_error import DAVError # type: ignore
from wsgidav.dav_provider import DAVCollection, DAVNonCollection, DAVProvider # type: ignore
from yaml import load, Loader

from .lrudict import LRUDict


def make_application(
    dbfile: bytes | str | PathLike, 
    config_path: str | Path = "", 
    cookies_path: str | Path = "", 
    strm_origin: bytes | str = "", 
    predicate: None | Callable[[MappingPath], bool] = None, 
    strm_predicate: None | Callable[[MappingPath], bool] = None, 
) -> Application:
    if isinstance(strm_origin, str):
        strm_origin = strm_origin.encode("utf-8")
    if config_path:
        config = load(open(config_path, encoding="utf-8"), Loader=Loader)
    else:
        config = {"simple_dc": {"user_mapping": {"*": True}}}
    if cookies_path:
        cookies_path = Path(cookies_path)
    else:
        cookies_path = Path("115-cookies.txt")
    client = P115Client(cookies_path, app="alipaymini", check_for_relogin=True)
    urlopen = partial(PoolManager(num_pools=128).request, "GET", preload_content=False)

    CON: Connection
    CON_FILE: Connection
    FIELDS = ("id", "name", "path", "ctime", "mtime", "sha1", "size", "pickcode", "is_dir")
    ROOT = {"id": 0, "name": "", "path": "/", "ctime": 0, "mtime": 0, "size": 0, "pickcode": "", "is_dir": 1}
    if strm_predicate:
        STRM_CACHE: LRUDict = LRUDict(65536)
    WRITE_LOCK = Lock()

    class DavPathBase:

        def __getattr__(self, attr, /):
            try:
                return self.attr[attr]
            except KeyError as e:
                raise AttributeError(attr) from e

        @locked_cacheproperty
        def creationdate(self, /) -> float:
            return self.ctime

        @locked_cacheproperty
        def ctime(self, /) -> float:
            return self.attr["ctime"]

        @locked_cacheproperty
        def mtime(self, /) -> float:
            return self.attr["mtime"]

        @locked_cacheproperty
        def name(self, /) -> str:
            return self.attr["name"]

        def get_creation_date(self, /) -> float:
            return self.ctime

        def get_display_name(self, /) -> str:
            return self.name

        def get_etag(self, /) -> str:
            return "%s-%s-%s" % (
                self.attr["pickcode"], 
                self.mtime, 
                self.size, 
            )

        def get_last_modified(self, /) -> float:
            return self.mtime

        def is_link(self, /) -> bool:
            return False

        def support_etag(self, /) -> bool:
            return True

        def support_modified(self, /) -> bool:
            return True

    class FileResource(DavPathBase, DAVNonCollection):

        def __init__(
            self, 
            /, 
            path: str, 
            environ: dict, 
            attr: dict, 
            is_strm: bool = False, 
        ):
            super().__init__(path, environ)
            self.attr = attr
            self.is_strm = is_strm
            if is_strm:
                STRM_CACHE[path] = self

        if strm_origin:
            @property
            def origin(self, /) -> bytes:
                return f"{self.environ['wsgi.url_scheme']}://{self.environ['HTTP_HOST']}".encode("utf-8")
        else:
            origin = strm_origin # type: ignore

        @property
        def size(self, /) -> int:
            if self.is_strm:
                return len(self.origin) + len(self.strm_data)
            return self.attr["size"]

        @locked_cacheproperty
        def strm_data(self, /) -> bytes:
            attr = self.attr
            name = encode_uri_component_loose(attr["name"])
            return bytes(f"/{name}?pickcode={attr['pickcode']}&id={attr['id']}&sha1={attr['sha1']}&size={attr['size']}", "utf-8")

        @property
        def url(self, /) -> str:
            scheme = self.environ["wsgi.url_scheme"]
            host = self.environ["HTTP_HOST"]
            return f"{scheme}://{host}?pickcode={self.attr['pickcode']}"

        def get_content(self, /):
            if self.is_strm:
                return BytesIO(self.origin + self.strm_data)
            fid = self.attr["id"]
            try:
                return CON_FILE.blobopen("data", "data", fid, readonly=True)
            except (OperationalError, SystemError):
                pass
            if self.attr["size"] >= 1024 * 64:
                raise DAVError(302, add_headers=[("Location", self.url)])
            CON_FILE.execute("""\
INSERT INTO data(id, data) VALUES(?, zeroblob(?)) 
ON CONFLICT(id) DO UPDATE SET data=excluded.data;""", (fid, self.attr["size"]))
            CON_FILE.commit()
            try:
                data = urlopen(self.url).read()
                with WRITE_LOCK:
                    with CON_FILE.blobopen("data", "data", fid) as fdst:
                        fdst.write(data)
                return CON_FILE.blobopen("data", "data", fid, readonly=True)
            except:
                CON_FILE.execute("DELETE FROM data WHERE id=?", (fid,))
                CON_FILE.commit()
                raise

        def get_content_length(self, /) -> int:
            return self.size

        def support_content_length(self, /) -> bool:
            return True

        def support_ranges(self, /) -> bool:
            return True

    class FolderResource(DavPathBase, DAVCollection):

        def __init__(
            self, 
            /, 
            path: str, 
            environ: dict, 
            attr: dict, 
        ):
            if not path.endswith("/"):
                path += "/"
            super().__init__(path, environ)
            self.attr = attr

        @locked_cacheproperty
        def children(self, /) -> dict[str, FileResource | FolderResource]:
            sql = """\
SELECT id, name, path, ctime, mtime, sha1, size, pickcode, is_dir
FROM data
WHERE parent_id = :id AND name NOT IN ('', '.', '..') AND name NOT LIKE '%/%';
"""
            children: dict[str, FileResource | FolderResource] = {}
            environ = self.environ
            for r in CON.execute(sql, self.attr):
                attr = dict(zip(FIELDS, r))
                is_strm = False
                name = attr["name"]
                path = attr["path"]
                if not attr["is_dir"] and strm_predicate and strm_predicate(MappingPath(attr)):
                    name = splitext(name)[0] + ".strm"
                    path = splitext(path)[0] + ".strm"
                    is_strm = True
                elif predicate and not predicate(MappingPath(attr)):
                    continue
                if attr["is_dir"]:
                    children[name] = FolderResource(path, environ, attr)
                else:
                    children[name] = FileResource(path, environ, attr, is_strm=is_strm)
            return children

        def get_descendants(
            self, 
            /, 
            collections: bool = True, 
            resources: bool = True, 
            depth_first: bool = False, 
            depth: Literal["0", "1", "infinity"] = "infinity", 
            add_self: bool = False, 
        ) -> list[FileResource | FolderResource]:
            descendants: list[FileResource | FolderResource] = []
            push = descendants.append
            if collections and add_self:
                push(self)
            if depth == "0":
                return descendants
            elif depth == "1":
                for item in self.children.values():
                    if item.attr["is_dir"]:
                        if collections:
                            push(item)
                    elif resources:
                        push(item)
                return descendants
            sql = """\
SELECT id, name, path, ctime, mtime, sha1, size, pickcode, is_dir
FROM data
WHERE path LIKE ? || '%' AND name NOT IN ('', '.', '..') AND name NOT LIKE '%/%'"""
            if collections and resources:
                pass
            elif collections:
                sql += " AND is_dir"
            elif resources:
                sql += " AND NOT is_dir"
            else:
                return descendants
            if depth_first:
                sql += "\nORDER BY path"
            else:
                sql += "\nORDER BY dirname(path)"
            environ = self.environ
            for r in CON.execute(sql, (self.path,)):
                attr = dict(zip(FIELDS, r))
                is_strm = False
                path = attr["path"]
                if not attr["is_dir"] and strm_predicate and strm_predicate(MappingPath(attr)):
                    path = splitext(path)[0] + ".strm"
                    is_strm = True
                elif predicate and not predicate(MappingPath(attr)):
                    continue
                if attr["is_dir"]:
                    push(FolderResource(path, environ, attr))
                else:
                    push(FileResource(path, environ, attr, is_strm=is_strm))
            return descendants

        def get_member(self, /, name: str) -> FileResource | FolderResource:
            if res := self.children.get(name):
                return res
            raise DAVError(404, self.path + name)

        def get_member_list(self, /) -> list[FileResource | FolderResource]:
            return list(self.children.values())

        def get_member_names(self, /) -> list[str]:
            return list(self.children.keys())

        def get_property_value(self, /, name: str):
            if name == "{DAV:}getcontentlength":
                return 0
            elif name == "{DAV:}iscollection":
                return True
            return super().get_property_value(name)

    class ServeDBProvider(DAVProvider):

        def __init__(self, /, dbfile: bytes | str | PathLike):
            nonlocal CON, CON_FILE
            CON = connect("file:%s?mode=ro" % quote(fsdecode(dbfile)), uri=True, check_same_thread=False)
            CON.create_function("dirname", 1, dirname)
            dbpath = CON.execute("SELECT file FROM pragma_database_list() WHERE name='main';").fetchone()[0]
            CON_FILE = connect("%s-file%s" % splitext(dbpath), check_same_thread=False)
            CON_FILE.execute("PRAGMA journal_mode = WAL;")
            CON_FILE.execute("""\
CREATE TABLE IF NOT EXISTS data (
    id INTEGER NOT NULL PRIMARY KEY,
    data BLOB,
    temp_path TEXT
);""")

        def __del__(self, /):
            try:
                CON.close()
            except:
                pass
            try:
                CON_FILE.close()
            except:
                pass

        def get_resource_inst(
            self, 
            /, 
            path: str, 
            environ: dict, 
        ) -> FolderResource | FileResource:
            is_dir = path.endswith("/")
            path = "/" + path.strip("/")
            if strm_predicate:
                if strm := STRM_CACHE.get(path):
                    return strm
                if path.endswith(".strm") and not is_dir:
                    dir_, name = splitpath(path)
                    inst = self.get_resource_inst(dir_, environ)
                    if not isinstance(inst, FolderResource):
                        raise DAVError(404, path)
                    return inst.get_member(name)
            if path == "/":
                return FolderResource("/", environ, ROOT)
            sql = "SELECT id, name, path, ctime, mtime, sha1, size, pickcode, is_dir FROM data WHERE path = ? LIMIT 1"
            record = CON.execute(sql, (path,)).fetchone()
            if not record:
                if path.endswith(".strm"):
                    sql = "SELECT id, name, path, ctime, mtime, sha1, size, pickcode, is_dir FROM data WHERE path LIKE ? || '.%' AND NOT is_dir LIMIT 1"
                    record = CON.execute(sql, (path[:-5],)).fetchone()
                    if record:
                        attr = dict(zip(FIELDS, record))
                        attr["path"] = path
                        return FileResource(path, environ, attr, is_strm=True)
                raise DAVError(404, path)
            attr = dict(zip(FIELDS, record))
            is_strm = False
            if not attr["is_dir"] and strm_predicate and strm_predicate(MappingPath(attr)):
                is_strm = True
                path = splitext(path)[0] + ".strm"
            elif predicate and not predicate(MappingPath(attr)):
                raise DAVError(404, path)
            if attr["is_dir"]:
                return FolderResource(path, environ, attr)
            else:
                return FileResource(path, environ, attr, is_strm=is_strm)

        def is_readonly(self, /) -> bool:
            return True

    app = Application()
    use_gzip_compression(app)

    @app.on_middlewares_configuration
    def configure_forwarded_headers(app: Application):
        app.middlewares.insert(0, ForwardedHeadersMiddleware(accept_only_proxied_requests=False))

    @app.lifespan
    async def register_client(app: Application):
        async with ClientSession(follow_redirects=False) as client:
            app.services.register(ClientSession, instance=client)
            yield

    @app.router.route("/", methods=["GET", "HEAD"])
    async def index(request: Request, session: ClientSession, pickcode: str = ""):
        if pickcode:
            user_agent = (request.get_first_header(b"User-agent") or b"").decode("latin-1")
            resp = await client.download_url_app(
                pickcode, 
                headers={"User-Agent": user_agent}, 
                request=blacksheep_request, 
                session=session, 
                async_=True, 
            )
            if not resp["state"]:
                return resp, 500
            for fid, info in resp["data"].items():
                if not info["url"]:
                    return resp, 404
                return redirect(info["url"]["url"])
        else:
            return redirect("/d")

    @app.router.route("/", methods=[
        "POST", "PUT", "DELETE", "CONNECT", "OPTIONS", 
        "TRACE", "PATCH", "MKCOL", "COPY", "MOVE", "PROPFIND", 
        "PROPPATCH", "LOCK", "UNLOCK", "REPORT", "ACL", 
    ])
    async def redirect_to_dav():
        return redirect("/d")

    @app.router.route("/<path:path>", methods=["GET", "HEAD"])
    async def resolve_path(request: Request, session: ClientSession, pickcode: str = "", path: str = ""):
        if pickcode:
            return redirect(f"/?pickcode={pickcode}")
        else:
            return redirect(f"/d/{path}")

    @app.router.route("/<path:path>", methods=[
        "POST", "PUT", "DELETE", "CONNECT", "OPTIONS", 
        "TRACE", "PATCH", "MKCOL", "COPY", "MOVE", "PROPFIND", 
        "PROPPATCH", "LOCK", "UNLOCK", "REPORT", "ACL", 
    ])
    async def resolve_path_to_dav(path: str):
        return redirect(f"/d/{path}")

    config.update({
        "host": "0.0.0.0", 
        "host": 0, 
        "mount_path": "/d", 
        "provider_mapping": {"/": ServeDBProvider(dbfile)}, 
    })
    wsgidav_app = WsgiDAVApp(config)
    app.mount("/d", WSGIMiddleware(wsgidav_app, workers=128, send_queue_size=256))
    return app

# TODO: wsgidav 速度比较一般，我需要自己实现一个 webdav
# TODO: 目前 webdav 是只读的，之后需要支持写入和删除，写入小文件不会上传，因此需要一个本地的 id，如果路径上原来有记录，则替换掉此记录（删除记录，生成本地 id 的数据，插入数据）
# TODO: 如果需要写入文件，会先把数据存入临时文件，等到关闭文件，再自动写入数据库。如果文件未被修改，则忽略，如果修改了，就用我本地的id替代原来的数据
# TODO: 文件可以被 append 写，这时打开时，会先把数据库的数据写到硬盘，然后打开这个临时文件
# TODO: 实现自动排除空目录而不展示



#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = ["main"]
__doc__ = """\
    🛸 115 数据库 WebDAV 服务，请先用 updatedb.py 采集数据 ✈️
"""

from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from pathlib import Path

if __name__ == "__main__":  
    from sys import path

    path[0] = str(Path(__file__).parents[1])
    parser = ArgumentParser(description=__doc__, formatter_class=RawTextHelpFormatter)
else:
    from .init import subparsers

    parser = subparsers.add_parser("dav", description=__doc__, formatter_class=RawTextHelpFormatter)


def parse_args(argv: None | list[str] = None, /) -> Namespace:
    args = parser.parse_args(argv)
    if args.version:
        from p115servedb import __version__
        print(".".join(map(str, __version__)))
        raise SystemExit(0)
    return args


def main(argv: None | list[str] | Namespace = None, /):
    if isinstance(argv, Namespace):
        args = argv
    else:
        args = parse_args(argv)

    import re
    import uvicorn

    from p115servedb.component.dav import make_application
    from path_predicate import make_predicate

    if args.fast_strm:
        predicate = make_predicate("""(
    path.is_dir() or
    path.media_type.startswith("image/") or
    path.suffix.lower() in (".nfo", ".ass", ".ssa", ".srt", ".idx", ".sub", ".txt", ".vtt", ".smi")
)""", type="expr")
    elif predicate := args.predicate or None:
        predicate = make_predicate(predicate, {"re": re}, type=args.predicate_type)
    if args.fast_strm:
        strm_predicate = make_predicate("""(
    path.media_type.startswith(("video/", "audio/")) and
    path.suffix.lower() != ".ass"
)""", type="expr")
    elif strm_predicate := args.strm_predicate or None:
        strm_predicate = make_predicate(strm_predicate, {"re": re}, type=args.strm_predicate_type)

    app = make_application(
        args.dbfile, 
        config_path=args.config_path, 
        cookies_path=args.cookies_path, 
        strm_origin=args.strm_origin, 
        predicate=predicate, 
        strm_predicate=strm_predicate, 
    )
    if args.config:
        from orjson import loads
        kwargs = loads(args.config)
    else:
        kwargs = {}
    debug = args.debug
    if debug:
        getattr(app, "logger").level = 10
        app.show_error_details = True
        kwargs["reload"] = True
    kwargs["host"] = args.host
    if args.port:
        kwargs["port"] = args.port
    elif not kwargs.get("port"):
        from socket import create_connection
        def get_available_ip(start: int = 1024, stop: int = 65536) -> int:
            for port in range(start, stop):
                try:
                    with create_connection(("127.0.0.1", port), timeout=1):
                        pass
                except OSError:
                    return port
            raise RuntimeError("no available ports")
        kwargs["port"] = get_available_ip()
    kwargs.setdefault("proxy_headers", True)
    kwargs.setdefault("server_header", False)
    kwargs.setdefault("forwarded_allow_ips", "*")
    kwargs.setdefault("timeout_graceful_shutdown", 1)
    uvicorn.run(app, **kwargs)


parser.add_argument("-f", "--dbfile", required=True, help="数据库文件路径")
parser.add_argument("-c", "--config-path", help="""webdav 配置文件路径，采用 yaml 格式，如需样板文件，请阅读：

    https://wsgidav.readthedocs.io/en/latest/user_guide_configure.html#sample-wsgidav-yaml

""")
parser.add_argument("-cp", "--cookies-path", default="", help="cookies 文件保存路径，默认为当前工作目录下的 115-cookies.txt")
parser.add_argument("-o", "--strm-origin", default="", help="strm 所用的 302 服务地址，如果不传，则自动确定")
parser.add_argument("-p1", "--predicate", help="断言，当断言的结果为 True 时，文件或目录会被显示")
parser.add_argument(
    "-t1", "--predicate-type", default="ignore", 
    choices=("ignore", "ignore-file", "expr", "lambda", "stmt", "module", "file", "re"), 
    help="""断言类型，默认值为 'ignore'
    - ignore       （默认值）gitignore 配置文本（有多个时用空格隔开），在文件路径上执行模式匹配，匹配成功则断言为 False
                   NOTE: https://git-scm.com/docs/gitignore#_pattern_format
    - ignore-file  接受一个文件路径，包含 gitignore 的配置文本（一行一个），在文件路径上执行模式匹配，匹配成功则断言为 False
                   NOTE: https://git-scm.com/docs/gitignore#_pattern_format
    - expr         表达式，会注入一个名为 path 的 p115.P115Path 对象
    - lambda       lambda 函数，接受一个 p115.P115Path 对象作为参数
    - stmt         语句，当且仅当不抛出异常，则视为 True，会注入一个名为 path 的 p115.P115Path 对象
    - module       模块，运行后需要在它的全局命名空间中生成一个 check 或 predicate 函数用于断言，接受一个 p115.P115Path 对象作为参数
    - file         文件路径，运行后需要在它的全局命名空间中生成一个 check 或 predicate 函数用于断言，接受一个 p115.P115Path 对象作为参数
    - re           正则表达式，模式匹配，如果文件的名字匹配此模式，则断言为 True
""")
parser.add_argument("-p2", "--strm-predicate", help="strm 断言（优先级高于 -p1/--predicate），当断言的结果为 True 时，文件会被显示为带有 .strm 后缀的文本文件，打开后是链接")
parser.add_argument(
    "-t2", "--strm-predicate-type", default="filter", 
    choices=("filter", "filter-file", "expr", "lambda", "stmt", "module", "file", "re"), 
    help="""断言类型，默认值为 'filter'
    - filter       （默认值）gitignore 配置文本（有多个时用空格隔开），在文件路径上执行模式匹配，匹配成功则断言为 True
                   请参考：https://git-scm.com/docs/gitignore#_pattern_format
    - filter-file  接受一个文件路径，包含 gitignore 的配置文本（一行一个），在文件路径上执行模式匹配，匹配成功则断言为 True
                   请参考：https://git-scm.com/docs/gitignore#_pattern_format
    - expr         表达式，会注入一个名为 path 的 p115.P115Path 对象
    - lambda       lambda 函数，接受一个 p115.P115Path 对象作为参数
    - stmt         语句，当且仅当不抛出异常，则视为 True，会注入一个名为 path 的 p115.P115Path 对象
    - module       模块，运行后需要在它的全局命名空间中生成一个 check 或 predicate 函数用于断言，接受一个 p115.P115Path 对象作为参数
    - file         文件路径，运行后需要在它的全局命名空间中生成一个 check 或 predicate 函数用于断言，接受一个 p115.P115Path 对象作为参数
    - re           正则表达式，模式匹配，如果文件的名字匹配此模式，则断言为 True
""")
parser.add_argument("-fs", "--fast-strm", action="store_true", help="""快速实现 媒体筛选 和 虚拟 strm，此命令优先级较高，相当于命令行指定

    --strm-predicate-type expr \\
    --strm-predicate '(
        path.media_type.startswith(("video/", "audio/")) and
        path.suffix.lower() != ".ass"
    )' \\
    --predicate-type expr \\
    --predicate '(
        path.is_dir() or
        path.media_type.startswith("image/") or
        path.suffix.lower() in (".nfo", ".ass", ".ssa", ".srt", ".idx", ".sub", ".txt", ".vtt", ".smi")
    )'
""")
parser.add_argument("-H", "--host", default="0.0.0.0", help="ip 或 hostname，默认值：'0.0.0.0'")
parser.add_argument("-P", "--port", default=8000, type=int, help="端口号，默认值：8000")
parser.add_argument("-d", "--debug", action="store_true", help="启用 debug 模式（会输出更详细的信息）")
parser.add_argument("-C", "--config", help="将被作为 JSON 解析然后作为关键字参数传给 `uvicorn.run`")
parser.add_argument("-v", "--version", action="store_true", help="输出版本号")
parser.set_defaults(func=main)


if __name__ == "__main__":
    main()




#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = ["ServedbFuseOperations"]

from . import __fuse_monkey_patch

import errno
import logging

from collections.abc import Callable
from functools import partial
from itertools import count
from pathlib import Path
from posixpath import split as splitpath, splitext
from sqlite3 import connect
from stat import S_IFDIR, S_IFREG
from typing import Final, BinaryIO
from unicodedata import normalize

from fuse import FUSE, Operations # type: ignore
from httpfile import Urllib3FileReader
from p115client import P115Client
from path_predicate import MappingPath

from .log import logger
from .lrudict import LRUDict


FIELDS: Final = ("id", "name", "path", "ctime", "mtime", "sha1", "size", "pickcode", "is_dir")


# Learning: 
#   - https://www.stavros.io/posts/python-fuse-filesystem/
#   - https://thepythoncorner.com/posts/2017-02-27-writing-a-fuse-filesystem-in-python/
class ServedbFuseOperations(Operations):

    def __init__(
        self, 
        /, 
        dbfile, 
        cookies: None | str | Path = None, 
        predicate: None | Callable[[MappingPath], bool] = None, 
        strm_predicate: None | Callable[[MappingPath], bool] = None, 
        strm_origin: str = "http://localhost:8000", 
    ):
        self.con = connect(dbfile, check_same_thread=False)
        self.client = P115Client(cookies, check_for_relogin=True) if cookies else None
        self._log = partial(logger.log, extra={"instance": repr(self)})
        self.predicate = predicate
        self.strm_predicate = strm_predicate
        self.strm_origin = strm_origin
        self._next_fh: Callable[[], int] = count(1).__next__
        self._fh_to_file: dict[int, tuple[BinaryIO, bytes]] = {}
        self.cache: LRUDict = LRUDict(1024)
        self.normpath_map: dict[str, str] = {}

    def __del__(self, /):
        self.close()

    def close(self, /):
        self.con.close()
        if self.client:
            self.client.close()
        popitem = self._fh_to_file.popitem
        while True:
            try:
                _, (file, _) = popitem()
                if file is not None:
                    file.close()
            except KeyError:
                break
            except:
                pass

    def getattr(self, /, path: str, fh: int = 0, _rootattr={"st_mode": S_IFDIR | 0o555}) -> dict:
        self._log(logging.DEBUG, "getattr(path=\x1b[4;34m%r\x1b[0m, fh=%r)", path, fh)
        if path == "/":
            return _rootattr
        dir_, name = splitpath(normalize("NFC", path))
        try:
            dird = self.cache[dir_]
        except KeyError:
            try:
                self.readdir(dir_)
                dird = self.cache[dir_]
            except BaseException as e:
                self._log(
                    logging.WARNING, 
                    "file not found: \x1b[4;34m%s\x1b[0m, since readdir failed: \x1b[4;34m%s\x1b[0m\n  |_ \x1b[1;4;31m%s\x1b[0m: %s", 
                    path, dir_, type(e).__qualname__, e, 
                )
                raise OSError(errno.EIO, path) from e
        try:
            return dird[name]
        except KeyError as e:
            self._log(
                logging.WARNING, 
                "file not found: \x1b[4;34m%s\x1b[0m\n  |_ \x1b[1;4;31m%s\x1b[0m: %s", 
                path, type(e).__qualname__, e, 
            )
            raise FileNotFoundError(errno.ENOENT, path) from e

    def open(self, /, path: str, flags: int = 0) -> int:
        self._log(logging.INFO, "open(path=\x1b[4;34m%r\x1b[0m, flags=%r)", path, flags)
        return self._next_fh()

    def _open(self, path: str, /, start: int = 0):
        attr = self.getattr(path)
        if attr.get("_data") is not None:
            return None, attr["_data"]
        pickcode = attr["_attr"]["pickcode"]
        if client := self.client:
            file = client.open(client.download_url(pickcode), http_file_reader_cls=Urllib3FileReader)
        else:
            file = Urllib3FileReader(f"{self.strm_origin}?pickcode={pickcode}")
        if attr["st_size"] <= 2048:
            return None, file.read()
        if start == 0:
            preread = file.read(2048)
        else:
            preread = b""
        return file, preread

    def read(self, /, path: str, size: int, offset: int, fh: int = 0) -> bytes:
        self._log(logging.DEBUG, "read(path=\x1b[4;34m%r\x1b[0m, size=%r, offset=%r, fh=%r)", path, size, offset, fh)
        if not fh:
            return b""
        try:
            try:
                file, preread = self._fh_to_file[fh]
            except KeyError:
                file, preread = self._fh_to_file[fh] = self._open(path, offset)
            cache_size = len(preread)
            if file is None:
                return preread[offset:offset+size]
            elif offset < cache_size:
                if offset + size <= cache_size:
                    return preread[offset:offset+size]
                elif file is not None:
                    file.seek(cache_size)
                    return preread[offset:] + file.read(offset+size-cache_size)
            file.seek(offset)
            return file.read(size)
        except BaseException as e:
            self._log(
                logging.ERROR, 
                "can't read file: \x1b[4;34m%s\x1b[0m\n  |_ \x1b[1;4;31m%s\x1b[0m: %s", 
                path, type(e).__qualname__, e, 
            )
            raise OSError(errno.EIO, path) from e

    def readdir(self, /, path: str, fh: int = 0) -> list[str]:
        self._log(logging.DEBUG, "readdir(path=\x1b[4;34m%r\x1b[0m, fh=%r)", path, fh)
        predicate = self.predicate
        strm_predicate = self.strm_predicate
        strm_origin = self.strm_origin
        path = normalize("NFC", path)
        children: dict[str, dict] = {}
        self.cache[path] = children
        realpath = self.normpath_map.get(path, path)
        try:
            if realpath == "/":
                sql = "SELECT id, name, path, ctime, mtime, sha1, size, pickcode, is_dir FROM data WHERE parent_id = 0;"
                cur = self.con.execute(sql)
            else:
                sql = """\
SELECT id, name, path, ctime, mtime, sha1, size, pickcode, is_dir
FROM data
WHERE parent_id = (SELECT id FROM data WHERE path = ? LIMIT 1) AND name NOT IN ('', '.', '..') AND name NOT LIKE '%/%';
"""
                cur = self.con.execute(sql, (realpath,))
            for r in cur:
                attr = dict(zip(FIELDS, r))
                data = None
                size = attr.get("size") or 0
                name = attr["name"]
                path = attr["path"]
                isdir = attr["is_dir"]
                if not isdir and strm_predicate and strm_predicate(MappingPath(attr)):
                    data = f"{strm_origin}?pickcode={attr['pickcode']}".encode("utf-8")
                    size = len(data)
                    name = splitext(name)[0] + ".strm"
                    path = splitext(path)[0] + ".strm"
                elif predicate and not predicate(MappingPath(attr)):
                    continue
                normname = normalize("NFC", name)
                children[normname] = dict(
                    st_mode=(S_IFDIR if isdir else S_IFREG) | 0o555, 
                    st_size=size, 
                    st_ctime=attr["ctime"], 
                    st_mtime=attr["mtime"], 
                    _attr=attr, 
                    _data=data, 
                )
                normpath = normalize("NFC", path)
                if path != normpath:
                    self.normpath_map[normpath] = path
            return [".", "..", *children]
        except BaseException as e:
            self._log(
                logging.ERROR, 
                "can't readdir: \x1b[4;34m%s\x1b[0m\n  |_ \x1b[1;4;31m%s\x1b[0m: %s", 
                path, type(e).__qualname__, e, 
            )
            raise OSError(errno.EIO, path) from e

    def release(self, /, path: str, fh: int = 0):
        self._log(logging.DEBUG, "release(path=\x1b[4;34m%r\x1b[0m, fh=%r)", path, fh)
        if not fh:
            return
        try:
            file, _ = self._fh_to_file.pop(fh)
            if file is not None:
                file.close()
        except KeyError:
            pass
        except BaseException as e:
            self._log(
                logging.ERROR, 
                "can't release file: \x1b[4;34m%s\x1b[0m\n  |_ \x1b[1;4;31m%s\x1b[0m: %s", 
                path, type(e).__qualname__, e, 
            )
            raise OSError(errno.EIO, path) from e

    def run(self, /, *args, **kwds):
        return FUSE(self, *args, **kwds)

